import logging
from io import StringIO
from typing import Dict
from logilab.common.graph import ordered_nodes

import yams.types as yams_types


def serialize_to_python(schema: yams_types.Schema) -> str:
    out = StringIO()
    w = out.write
    w("from yams.buildobjs import *\n\n")

    graph: Dict = {}

    for entity in schema.entities():
        targets = graph.setdefault(entity, set())
        if entity.final:
            continue
        base = entity._specialized_type
        if base is None:
            continue
        if isinstance(base, str) and ", " in base:
            targets |= set(base.split(", "))
        else:
            targets.add(base)

    for e in reversed(ordered_nodes(graph)):
        if not e.final:
            if e._specialized_type:
                base = e._specialized_type
            else:
                base = "EntityType"
            w(f"class {e.type}({base}):\n")
            attr_defs = list(e.attribute_definitions())
            if attr_defs:
                for attr, obj in attr_defs:
                    w(f"    {attr.type} = {obj.type}()\n")
            else:
                w("    pass\n")
            w("\n")

    for r in schema.relations():
        if not r.final:
            if r.subjects() and r.objects():
                w(f"class {r.type}(RelationDefinition):\n")
                w("    subject = (%s,)\n" % ", ".join("'%s'" % x for x in r.subjects()))
                w("    object = (%s,)\n" % ", ".join("'%s'" % x for x in r.objects()))
                w("\n")
            else:
                logging.warning(f"relation definition {r.type} missing subject/object")

    return out.getvalue()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.

"""unit tests for module yams.diff"""

from logilab.common.testlib import TestCase, unittest_main

from yams.reader import build_schema_from_namespace
from yams.buildobjs import String, Int, Float, EntityType, SubjectRelation
from yams.diff import properties_from, schema_diff


class PersonBase(EntityType):
    nom = String()
    prenom = String()


class PersonAttrMod(EntityType):
    nom = String()
    prenom = Float()


class PersonAttrAdd(EntityType):
    nom = String()
    prenom = String()
    age = Int()
    is_friend_of = SubjectRelation("PersonAttrMod")


class PersonAttrAdd2(EntityType):
    nom = String()
    prenom = String()
    salaire = Float()
    is_friend_of = SubjectRelation("PersonAttrAdd")


class PersonAttrAdd3(EntityType):
    nom = String()
    prenom = String()
    age = Int()
    salaire = Float()
    is_spouse_of = SubjectRelation("PersonAttrAdd")


def create_schema_1():
    class Affaire(EntityType):
        nom = String()
        associate_affaire = SubjectRelation("Affaire")

    return build_schema_from_namespace(
        [("PersonBase", PersonBase), ("Affaire", Affaire), ("PersonAttrMod", PersonAttrMod)]
    )


def create_schema_2():
    class Affaire(EntityType):
        nom = String(maxsize=150)
        numero = Int()
        associate_affaire = SubjectRelation("Affaire", cardinality="**")
        associate_person = SubjectRelation("PersonBase")

    return build_schema_from_namespace([("PersonBase", PersonBase), ("Affaire", Affaire)])


class PropertiesFromTC(TestCase):
    expected_default_attr_perms = (
        " {\n\t\t\t'read': ('guests','managers','users'),"
        "\n\t\t\t'add': ('managers','users'),"
        "\n\t\t\t'update': ('managers','owners')\n\t\t}"
    )

    def build_relation_definition(self, props_ref):
        class AType(EntityType):
            attr = String(**props_ref)

        schema = build_schema_from_namespace(locals().items())
        return schema["AType"].relation_definition("attr")

    def build_props_dict(self, props_ref):
        result = {
            "default": None,
            "description": "",
            "fulltextindexed": False,
            "indexed": False,
            "internationalizable": False,
            "order": 1,
            "required": False,
            "uid": False,
        }
        result.update(props_ref)
        return result

    def test_properties_from_final_attributes_1(self):
        props_ref = {"required": True, "default": "toto", "description": "something"}
        self.assertEqual(
            {
                "default": "toto",
                "required": True,
                "__permissions__": self.expected_default_attr_perms,
                "description": "'something'",
                "order": 1,
            },
            properties_from(self.build_relation_definition(props_ref)),
        )

    def test_properties_from_final_attributes_2(self):
        props_ref = {}
        self.assertEqual(
            {"__permissions__": self.expected_default_attr_perms, "order": 1},
            properties_from(self.build_relation_definition(props_ref)),
        )

    def test_properties_from_final_attributes_3(self):
        props_ref = {"default": None, "required": False}
        self.assertEqual(
            {"__permissions__": self.expected_default_attr_perms, "order": 1},
            properties_from(self.build_relation_definition(props_ref)),
        )

    def test_constraint_properties_1(self):
        props_ref = {"maxsize": 150, "required": False}
        self.assertEqual(
            {"maxsize": 150, "__permissions__": self.expected_default_attr_perms, "order": 1},
            properties_from(self.build_relation_definition(props_ref)),
        )

    def test_constraint_properties_2(self):
        props_ref = {"unique": True, "required": False}
        self.assertEqual(
            {"__permissions__": self.expected_default_attr_perms, "unique": True, "order": 1},
            properties_from(self.build_relation_definition(props_ref)),
        )

    def test_constraint_properties_3(self):
        props_ref = {"vocabulary": ("aaa", "bbbb", "ccccc"), "required": False, "maxsize": 20}
        relation_definition = self.build_relation_definition(props_ref)
        props_ref["maxsize"] = 5
        self.assertEqual(
            {
                "maxsize": 5,
                "__permissions__": self.expected_default_attr_perms,
                "order": 1,
                "vocabulary": ["aaa", "bbbb", "ccccc"],
            },
            properties_from(relation_definition),
        )

    def test_permissions(self):
        props_ref = {
            "__permissions__": {
                "add": ("managers", "users"),
                "update": ("owners", "managers"),
                "read": ("users", "managers", "guests"),
            },
        }
        relation_definition = self.build_relation_definition(props_ref)
        self.assertEqual(
            {
                "__permissions__": self.expected_default_attr_perms,
                "order": 1,
            },
            properties_from(relation_definition),
        )


class SchemaDiff(TestCase):
    def test_schema_diff(self):
        schema1 = create_schema_1()
        schema2 = create_schema_2()
        output1, output2 = schema_diff(schema1, schema2)
        self.assertMultiLineEqual(open(output1).read(), open(self.datapath("schema1.txt")).read())
        self.assertMultiLineEqual(open(output2).read(), open(self.datapath("schema2.txt")).read())


if __name__ == "__main__":
    unittest_main()



# copyright 2004-2025 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <https://www.gnu.org/licenses/>.
"""Classes used to build a schema."""

from typing import Optional, List
from warnings import warn
from copy import copy
from difflib import get_close_matches

from typing import Any, Generator, Tuple, Type, Union, Dict, Sequence, Iterable, Set, cast, TypeVar

from logilab.common import attrdict, deprecation, nullobject

import yams.types as yams_types

from yams import (
    BASE_TYPES,
    MARKER,
    BadSchemaDefinition,
    KNOWN_METAATTRIBUTES,
    DEFAULT_ETYPEPERMS,
    DEFAULT_RELPERMS,
    DEFAULT_ATTRPERMS,
    DEFAULT_COMPUTED_ATTRPERMS,
)
from yams.constraints import (
    SizeConstraint,
    UniqueConstraint,
    BaseConstraint,
    StaticVocabularyConstraint,
    FORMAT_CONSTRAINT,
)
from yams.schema import RelationDefinitionSchema

Defined = Dict[Union[str, Tuple[str, str, str]], Union["autopackage", "Definition"]]

__docformat__: str = "restructuredtext en"

# will be modified by the yams'reader when schema is
# beeing read
PACKAGE: str = "<builtin>"


__all__: Tuple[str, ...] = (
    "EntityType",
    "RelationType",
    "RelationDefinition",
    "SubjectRelation",
    "ObjectRelation",
    "RichString",
) + tuple(BASE_TYPES)

# EntityType properties
ETYPE_PROPERTIES: Tuple[str, ...] = ("description", "__permissions__", "__unique_together__")

# RelationType properties. Don't put description inside, handled specifically
RTYPE_PROPERTIES: Tuple[str, ...] = ("symmetric", "inlined", "fulltext_container")

# RelationDefinition properties have to be computed dynamically since new ones
# may be added at runtime


def _RELATION_DEFINITION_PROPERTIES() -> Tuple[str, ...]:
    base: Set[str] = RelationDefinitionSchema.ALL_PROPERTIES()

    # infered is an internal property and should not be specified explicitly
    base.remove("infered")

    # replace permissions by __permissions__ as it's spelled that way in schema
    # definition files
    base.remove("permissions")
    base.add("__permissions__")

    return tuple(base)


_RDEF_PROPERTIES = deprecation.callable_renamed(
    old_name="_RDEF_PROPERTIES", new_function=_RELATION_DEFINITION_PROPERTIES, version="0.48"
)


# regroup all rtype/rdef properties as they may be defined one on each other in
# some cases


def _RELATION_PROPERTIES() -> Tuple[str, ...]:
    return RTYPE_PROPERTIES + _RELATION_DEFINITION_PROPERTIES()


_REL_PROPERTIES = deprecation.callable_renamed(
    old_name="_REL_PROPERTIES", new_function=_RELATION_PROPERTIES, version="0.48"
)


# pre 0.37 backward compat
RDEF_PROPERTIES = ()  # stuff added here is also added to underlying dict, nevermind


CREATION_RANK: int = 0


def _add_constraint(kwargs: Dict[str, Any], constraint: BaseConstraint) -> None:
    """Add constraint to param kwargs."""
    constraints: List[BaseConstraint] = kwargs.setdefault("constraints", [])

    for number, existing_constraint in enumerate(constraints):
        if existing_constraint.__class__ is constraint.__class__:
            constraints[number] = constraint

            return

    constraints.append(constraint)


@deprecation.argument_renamed(old_name="rdef", new_name="relation_definition", version="0.48")
@deprecation.argument_renamed(old_name="insertidx", new_name="insert_index", version="0.48")
def _add_relation(
    relations: List["RelationDefinition"],
    relation_definition: "RelationDefinition",
    name: Optional[yams_types.DefinitionName] = None,
    insert_index: Optional[int] = None,
) -> None:
    """Add relation (param relation_definition) to list of relations (param relations)."""
    if name is not None:
        relation_definition.name = name

    if insert_index is None:
        insert_index = len(relations)

    cast(int, insert_index)

    relations.insert(insert_index, relation_definition)

    if getattr(relation_definition, "metadata", {}):
        # mypy: "RelationDefinition" has no attribute "metadata"
        # dynamic attribute tested in the if
        for meta_name, value in relation_definition.metadata.items():  # type: ignore
            assert meta_name in KNOWN_METAATTRIBUTES

            insert_index += 1  # insert meta after main
            meta_rel_name: str = "_".join(((name or relation_definition.name), meta_name))

            _add_relation(relations, value, meta_rel_name, insert_index)


_KwargsKeyType = TypeVar("_KwargsKeyType")


def _check_kwargs(kwargs: Dict[_KwargsKeyType, Any], attributes: Sequence[_KwargsKeyType]) -> None:
    """Check that all keys of kwargs are actual attributes."""
    for key in kwargs:
        if key not in attributes:
            if key == "inline" and "inlined" in attributes:
                raise BadSchemaDefinition(
                    '"inline" is not a possible property name, you probably want "inlined"'
                )
            message = f"no such property {key!r} in {attributes!r}"
            if len(key) > 3:
                matches = get_close_matches(key, attributes)
                nb_matches = len(matches)
                if nb_matches == 1:
                    attribute = matches[0]
                    message += f"\nyou probably wanted property {attribute!r}"
                elif nb_matches > 1:
                    attributes = matches.join(", ")
                    message += f"\nyou probably wanted one of this properties: {attributes!r}"
            raise BadSchemaDefinition(message)


@deprecation.argument_renamed(old_name="fromobj", new_name="from_object", version="0.48")
@deprecation.argument_renamed(old_name="toobj", new_name="to_object", version="0.48")
def _copy_attributes(from_object: Any, to_object: Any, attributes: Iterable[str]) -> None:
    for attribute in attributes:
        value = getattr(from_object, attribute, MARKER)

        if value is MARKER:
            continue

        object_value = getattr(to_object, attribute, MARKER)

        if object_value is not MARKER and value != object_value:
            relation_name = getattr(to_object, "name", None) or to_object.__name__

            raise BadSchemaDefinition(
                "conflicting values %r/%r for property %s of relation %r"
                % (object_value, value, attribute, relation_name)
            )

        setattr(to_object, attribute, value)


def register_base_types(schema: yams_types.Schema) -> None:
    """add base (final) entity types to the given schema"""
    for entity_type in BASE_TYPES:
        entity_definition = EntityType(name=entity_type)

        schema.add_entity_type(entity_definition)


# first class schema definition objects #######################################


class autopackage(type):
    def __new__(
        mcs: "Type[autopackage]", name: str, bases: Tuple, classdict: Dict[str, Any]
    ) -> Any:
        classdict["package"] = PACKAGE

        return super(autopackage, mcs).__new__(mcs, name, bases, classdict)


class Definition(metaclass=autopackage):
    """Abstract class for entity / relation definition classes."""

    meta = MARKER
    description: Union[nullobject, str, None] = MARKER
    __permissions__ = MARKER

    def __init__(self, name=None) -> None:
        self.name: yams_types.DefinitionName = yams_types.DefinitionName(
            name or getattr(self, "name", None) or self.__class__.__name__
        )

        if self.__doc__:
            self.description = " ".join(self.__doc__.split())
            cast(str, self.description)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r} @{id(self):x}>"

    @classmethod
    def expand_type_definitions(cls: Type["Definition"], defined: Defined) -> None:
        """Schema building step 1: register definition objects by adding them
        to the `defined` dictionnary.
        """
        raise NotImplementedError()

    @classmethod
    def expand_relation_definitions(
        cls: Type["Definition"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """Schema building step 2: register all relations definition,
        expanding wildcard if necessary.
        """
        raise NotImplementedError()

    def get_permissions(self, final: bool = False) -> yams_types.Permissions:
        if self.__permissions__ is MARKER:
            if final:
                return DEFAULT_ATTRPERMS

            return DEFAULT_RELPERMS

        return self.__permissions__

    @classmethod
    @deprecation.argument_renamed(old_name="perms", new_name="permissions", version="0.48")
    def set_permissions(cls: Type["Definition"], permissions: yams_types.Permissions) -> None:
        cls.__permissions__ = permissions


# classes used to define relationships within entity type classes ##################


# has to be defined before the metadefinition metaclass which "isinstance" this
# class
@deprecation.attribute_renamed(old_name="etype", new_name="entity_type", version="0.48")
class ObjectRelation:
    __permissions__ = MARKER
    cardinality = MARKER
    constraints = MARKER

    @deprecation.argument_renamed(old_name="etype", new_name="entity_type", version="0.48")
    def __init__(
        self, entity_type: Union[str, Tuple[str, ...]], override: bool = False, **kwargs
    ) -> None:
        if self.__class__.__name__ == "ObjectRelation":
            warn(
                "[yams 0.29] ObjectRelation is deprecated, use RelationDefinition subclass",
                DeprecationWarning,
                stacklevel=2,
            )

        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank: int = CREATION_RANK

        self.package: str = PACKAGE
        self.name: str = "<undefined>"
        self.entity_type: Union[str, Tuple[str]] = entity_type
        self.override: bool = override

        if self.constraints:
            self.constraints = list(self.constraints)

        if kwargs.pop("meta", None):
            warn("[yams 0.37.0] meta is deprecated", DeprecationWarning, stacklevel=3)

        try:
            _check_kwargs(kwargs, _RELATION_PROPERTIES())
        except BadSchemaDefinition as bad:
            # XXX (auc) bad field name + required attribute can lead there
            # instead of schema.py ~ 920
            bad_schema_definition = BadSchemaDefinition(
                "%s in relation to entity %r (also is %r defined ? "
                "(check two lines above in the backtrace))" % (bad.args, entity_type, entity_type)
            )
            # mypy: "BadSchemaDefinition" has no attribute "tb_offset"
            # hack to transport information
            bad_schema_definition.tb_offset = 2  # type: ignore

            raise bad_schema_definition

        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return f"{self.__dict__['name']} {self.__dict__['entity_type']}"


class SubjectRelation(ObjectRelation):
    uid = MARKER
    indexed = MARKER
    fulltextindexed = MARKER
    internationalizable = MARKER
    default = MARKER

    def __repr__(self) -> str:
        return f"{self.__dict__['entity_type']} {self.__dict__['name']}"


class AbstractTypedAttribute(SubjectRelation):
    """AbstractTypedAttribute is not directly instantiable

    subclasses must provide a <entity_type> attribute to be instantiable
    """

    def __init__(
        self,
        metadata: Optional[Dict[str, "AbstractTypedAttribute"]] = None,
        required: bool = False,
        maxsize: Optional[int] = None,
        formula=MARKER,
        vocabulary: Optional[List[str]] = None,
        unique: Optional[bool] = None,
        override: bool = False,
        **kwargs,
    ):
        # Store metadata
        if metadata is None:
            metadata = {}
        self.metadata: Dict[Any, "AbstractTypedAttribute"] = metadata

        # transform "required" into "cardinality"
        if required:
            cardinality = "11"
        else:
            cardinality = "?1"
        kwargs["cardinality"] = cardinality

        # transform maxsize into SizeConstraint
        assert maxsize is None or maxsize >= 0
        if maxsize is not None:
            _add_constraint(kwargs, SizeConstraint(max=maxsize))

        # formula
        self.formula = formula

        # transform vocabulary into StaticVocabularyConstraint
        if vocabulary is not None:
            self.set_vocabulary(vocabulary, kwargs)

        # transform unique into UniqueConstraint
        if unique:
            _add_constraint(kwargs, UniqueConstraint())

        # use the entity_type attribute provided by subclasses
        kwargs["override"] = override
        super().__init__(self.entity_type, **kwargs)

        # reassign creation rank
        #
        # Main attribute are marked as created before it's metadata.
        # order in meta data is preserved.
        if self.metadata:
            meta: List[AbstractTypedAttribute] = sorted(
                metadata.values(), key=lambda x: x.creation_rank
            )

            if meta[0].creation_rank < self.creation_rank:
                _previous: AbstractTypedAttribute = self

                for _next in meta:
                    if _previous.creation_rank < _next.creation_rank:
                        break

                    _previous.creation_rank, _next.creation_rank = (
                        _next.creation_rank,
                        _previous.creation_rank,
                    )
                    _next = _previous

    def set_vocabulary(self, vocabulary: List[str], kwargs=None) -> None:
        if kwargs is None:
            kwargs = self.__dict__

        # constraints = kwargs.setdefault('constraints', [])
        _add_constraint(kwargs, StaticVocabularyConstraint(vocabulary))

        if self.__class__.__name__ == "String":  # XXX
            max_size = max(len(x) for x in vocabulary)

            _add_constraint(kwargs, SizeConstraint(max=max_size))

    def __repr__(self) -> str:
        return f"<{self.__dict__['name']}({self.__dict__['entity_type']})>"


@deprecation.argument_renamed(old_name="etype", new_name="entity_type", version="0.48")
def make_type(entity_type: str) -> Type[AbstractTypedAttribute]:
    """create a python class for a Yams base type.

    Notice it is now possible to create a specific type with user-defined
    behaviour, e.g.:

        Geometry = make_type('Geometry') # (c.f. postgis)

    will allow the use of:

        Geometry(geom_type='POINT')

    in a Yams schema, provided in this example that `geom_type` is specified to
    the :func:`yams.register_base_type` function which should be called prior to
    make_type.
    """
    assert entity_type in BASE_TYPES
    return type(entity_type, (AbstractTypedAttribute,), {"entity_type": entity_type})


# build a specific class for each base type
class String(AbstractTypedAttribute):
    entity_type: str = "String"


class Password(AbstractTypedAttribute):
    entity_type: str = "Password"


class Bytes(AbstractTypedAttribute):
    entity_type: str = "Bytes"


class Int(AbstractTypedAttribute):
    entity_type: str = "Int"


class BigInt(AbstractTypedAttribute):
    entity_type: str = "BigInt"


class Float(AbstractTypedAttribute):
    entity_type: str = "Float"


class Boolean(AbstractTypedAttribute):
    entity_type: str = "Boolean"


class Decimal(AbstractTypedAttribute):
    entity_type: str = "Decimal"


class Time(AbstractTypedAttribute):
    entity_type: str = "Time"


class Date(AbstractTypedAttribute):
    entity_type: str = "Date"


class Datetime(AbstractTypedAttribute):
    entity_type: str = "Datetime"


class TZTime(AbstractTypedAttribute):
    entity_type: str = "TZTime"


class TZDatetime(AbstractTypedAttribute):
    entity_type: str = "TZDatetime"


class Interval(AbstractTypedAttribute):
    entity_type: str = "Interval"


# provides a RichString factory for convenience
def RichString(
    default_format: str = "text/plain",
    format_constraints: Optional[List[BaseConstraint]] = None,
    required: bool = False,
    maxsize: Optional[int] = None,
    formula=MARKER,
    vocabulary: Optional[List[str]] = None,
    unique: Optional[bool] = None,
    override: bool = False,
    **kwargs,
):
    """RichString is a convenience attribute type for attribute containing text
    in a format that should be specified in another attribute.

    The following declaration::

      class Card(EntityType):
          content = RichString(fulltextindexed=True, default_format='text/rest')

    is equivalent to::

      class Card(EntityType):
          content_format = String(internationalizable=True,
                                  default='text/rest', constraints=[FORMAT_CONSTRAINT])
          content  = String(fulltextindexed=True)
    """
    format_args = {"default": default_format, "maxsize": 50}

    if format_constraints is None:
        format_args["constraints"] = [FORMAT_CONSTRAINT]
    else:
        format_args["constraints"] = format_constraints

    # mypy: Argument 2 to "String" has incompatible type "**Dict[str, object]"; expected
    # mypy: "Optional[bool]"
    # really looks like mypy is failing on AbstractTypedAttribute constructor here
    meta: Dict[str, AbstractTypedAttribute] = {
        "format": String(internationalizable=True, **format_args)  # type: ignore
    }

    return String(
        metadata=meta,
        required=required,
        maxsize=maxsize,
        formula=formula,
        vocabulary=vocabulary,
        unique=unique,
        override=override,
        **kwargs,
    )


# other schema definition classes ##############################################


class metadefinition(autopackage):
    """Metaclass that builds the __relations__ attribute of EntityType's
    subclasses.
    """

    stacklevel = 3

    def __new__(
        mcs: "Type[metadefinition]", name: str, bases: Tuple, classdict: Dict[str, Any]
    ) -> Any:
        # Move (any) relation from the class dict to __relations__ attribute
        relations_list: List = classdict.setdefault("__relations__", [])
        relations: Dict[str, Any] = dict(
            (relation_definition.name, relation_definition)
            for relation_definition in relations_list
        )

        for relation_name, relation_definition in list(classdict.items()):
            if isinstance(relation_definition, ObjectRelation):
                # relation's name **must** be removed from class namespace
                # to avoid conflicts with instance's potential attributes
                del classdict[relation_name]

                relations[relation_name] = relation_definition

        # handle logical inheritance
        if "__specializes_schema__" in classdict:
            specialized = bases[0]
            classdict["__specializes__"] = specialized.__name__

            if "__specialized_by__" not in specialized.__dict__:
                specialized.__specialized_by__ = []

            specialized.__specialized_by__.append(name)

        # Initialize processed class
        class_definition = super(metadefinition, mcs).__new__(mcs, name, bases, classdict)

        for relation_name, relation_definition in relations.items():
            _add_relation(class_definition.__relations__, relation_definition, relation_name)

        # take base classes'relations into account
        for base in bases:
            for relation_definition in getattr(base, "__relations__", ()):
                if (
                    relation_definition.name not in relations
                    or not relations[relation_definition.name].override
                ):
                    if isinstance(relation_definition, RelationDefinition):
                        relation_definition = copy(relation_definition)

                        if relation_definition.subject == base.__name__:
                            relation_definition.subject = name

                        if relation_definition.object == base.__name__:
                            relation_definition.object = name

                    relations_list.append(relation_definition)
                else:
                    relations[relation_definition.name].creation_rank = (
                        relation_definition.creation_rank
                    )

        # sort relations by creation rank
        class_definition.__relations__ = sorted(relations_list, key=lambda r: r.creation_rank)

        return class_definition


class EntityType(Definition, metaclass=metadefinition):
    # :FIXME reader magic forbids to define a docstring...
    #  an entity has attributes and can be linked to other entities by
    #  relations. Both entity attributes and relationships are defined by
    #  class attributes.
    #
    #  kwargs keys must have values in ETYPE_PROPERTIES
    #
    #  Example:
    #
    #  >>> class Project(EntityType):
    #  ...     name = String()
    #  >>>
    #
    #  After instanciation, EntityType can we altered with dedicated class methods:
    #
    #  .. currentmodule:: yams.buildobjs
    #
    #   .. automethod:: EntityType.extend
    #   .. automethod:: EntityType.add_relation
    #   .. automethod:: EntityType.insert_relation_after
    #   .. automethod:: EntityType.remove_relation
    #   .. automethod:: EntityType.get_relation
    #   .. automethod:: EntityType.get_relations

    __permissions__: yams_types.Permissions = DEFAULT_ETYPEPERMS

    def __init__(self, name: Optional[str] = None, **kwargs) -> None:
        super().__init__(name)

        _check_kwargs(kwargs, ETYPE_PROPERTIES)

        self.__dict__.update(kwargs)
        self.specialized_type: Optional[str] = self.__class__.__dict__.get("__specializes__")

    def __str__(self) -> str:
        return f"entity type {self.name!r}"

    @property
    def specialized_by(self) -> List[str]:
        return self.__class__.__dict__.get("__specialized_by__", [])

    @classmethod
    def expand_type_definitions(cls: Type["EntityType"], defined: Defined) -> None:
        """Schema building step 1: register definition objects by adding
        them to the `defined` dictionnary.
        """
        name: str = getattr(cls, "name", cls.__name__)

        assert cls is not defined.get(name), f"duplicate registration: {name}"
        assert (
            name not in defined
        ), "type '%s' was already defined here %s, new definition here %s" % (
            name,
            defined[name].__module__,
            cls,
        )

        # mypy: "Type[EntityType]" has no attribute "_defined"
        # dynamic attribute
        # XXX may be used later (eg .add_relation())
        cls._defined: Defined = defined  # type: ignore

        defined[name] = cls

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation in cls.__relations__:  # type: ignore
            cls._ensure_relation_type(relation)

    @classmethod
    def _ensure_relation_type(cls: Type["EntityType"], relation: ObjectRelation) -> bool:
        """Check the type the relation

        return False if the class is not yet finalized
        (XXX raise excep instead ?)"""

        relation_type = RelationType(relation.name)

        _copy_attributes(relation, relation_type, RTYPE_PROPERTIES)

        # assert hasattr(cls, '_defined'), "Type definition for %s not yet expanded.
        # you can't register new type through it" % cls

        if not hasattr(cls, "_defined"):
            return False

        # mypy: "Type[EntityType]" has no attribute "_defined"
        # dynamically set attribute
        defined = cls._defined  # type: ignore

        if relation.name in defined:
            _copy_attributes(relation_type, defined[relation.name], RTYPE_PROPERTIES)
        else:
            defined[relation.name] = relation_type

        return True

    @classmethod
    def expand_relation_definitions(
        cls: Type["EntityType"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcards if necessary
        """
        order: int = 1
        name: str = getattr(cls, "name", cls.__name__)
        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation in cls.__relations__:  # type: ignore
            if isinstance(relation, SubjectRelation):
                relation_definition = RelationDefinition(
                    subject=name,
                    name=relation.name,
                    object=relation.entity_type,
                    order=order,
                    package=relation.package,
                )

                _copy_attributes(relation, relation_definition, relation_definitions_properties)

            elif isinstance(relation, ObjectRelation):
                relation_definition = RelationDefinition(
                    subject=relation.entity_type,
                    name=relation.name,
                    object=name,
                    order=order,
                    package=relation.package,
                )

                _copy_attributes(relation, relation_definition, relation_definitions_properties)

            elif isinstance(relation, RelationDefinition):
                relation_definition = relation
            else:
                raise BadSchemaDefinition(f"dunno how to handle {relation}")

            order += 1
            relation_definition._add_relations(defined, schema)

    # methods that can be used to extend an existant schema definition ########

    @classmethod
    @deprecation.attribute_renamed(
        old_name="othermetadefcls", new_name="other_meta_definition_class", version="0.48"
    )
    def extend(cls: Type["EntityType"], other_meta_definition_class) -> None:
        """add all relations of ``other_meta_definition_class`` to the current class"""
        for relation_definition in other_meta_definition_class.__relations__:
            cls.add_relation(relation_definition)

    @classmethod
    @deprecation.argument_renamed(old_name="rdef", new_name="relation_definition", version="0.48")
    def add_relation(
        cls: Type["EntityType"], relation_definition: ObjectRelation, name: Optional[str] = None
    ) -> None:
        """Add ``relation_definition`` relation to the class"""
        if name:
            relation_definition.name = name

        if cls._ensure_relation_type(relation_definition):
            # mypy: "Type[EntityType]" has no attribute "__relations__"
            # dynamically set attribute, full yams magic
            _add_relation(cls.__relations__, relation_definition, name)  # type: ignore

            # mypy: "Type[EntityType]" has no attribute "_defined"
            # dynamically set attribute
            if (
                getattr(relation_definition, "metadata", {})
                and relation_definition not in cls._defined  # type: ignore
            ):
                for meta_name in relation_definition.metadata:  # type: ignore
                    format_attr_name = "_".join(((name or relation_definition.name), meta_name))
                    relation_definition = next(cls.get_relations(format_attr_name))
                    cls._ensure_relation_type(relation_definition)

        else:
            # mypy: "Type[EntityType]" has no attribute "__relations__"
            # dynamically set attribute, full yams magic
            _add_relation(cls.__relations__, relation_definition, name=name)  # type: ignore

    @classmethod
    @deprecation.argument_renamed(old_name="rdef", new_name="relation_definition", version="0.48")
    @deprecation.argument_renamed(
        old_name="afterrelname", new_name="after_relation_name", version="0.48"
    )
    def insert_relation_after(
        cls: Type["EntityType"],
        after_relation_name: str,
        name: str,
        relation_definition: ObjectRelation,
    ) -> None:
        """Add ``relation_definition`` relation to the class right after another"""
        # FIXME change order of arguments to relation_definition, name, after_relation_name ?
        relation_definition.name = name
        cls._ensure_relation_type(relation_definition)

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        number = 0
        for number, rel in enumerate(cls.__relations__):  # type: ignore
            if rel.name == after_relation_name:
                break

        else:
            raise BadSchemaDefinition(f"can't find {after_relation_name} relation on {cls}")

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        _add_relation(cls.__relations__, relation_definition, name, number + 1)  # type: ignore

    @classmethod
    def remove_relation(cls: Type["EntityType"], name: str) -> None:
        """Remove relation from the class"""

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation_definition in cls.get_relations(name):
            cls.__relations__.remove(relation_definition)  # type: ignore

    @classmethod
    def get_relations(cls: Type["EntityType"], name: str) -> Generator[ObjectRelation, Any, None]:
        """Iterate over relations definitions that match the ``name`` parameters

        It may iterate multiple definitions when the class is both object and
        sujet of a relation:
        """

        # mypy: "Type[EntityType]" has no attribute "__relations__"
        # dynamically set attribute, full yams magic
        for relation_definition in cls.__relations__[:]:  # type: ignore
            if relation_definition.name == name:
                yield relation_definition

    @classmethod
    def get_relation(cls: Type["EntityType"], name: str) -> ObjectRelation:
        """Return relation definitions by name. Fails if there is multiple one."""
        relations: Tuple[ObjectRelation, ...] = tuple(cls.get_relations(name))

        assert len(relations) == 1, "can't use get_relation for relation with multiple definitions"

        return relations[0]


class RelationType(Definition):
    symmetric = MARKER
    inlined = MARKER
    fulltext_container = MARKER
    rule = MARKER

    def __init__(self, name: Optional[str] = None, **kwargs) -> None:
        """kwargs must have values in RTYPE_PROPERTIES"""
        super().__init__(name)

        if kwargs.pop("meta", None):
            warn("[yams 0.37] meta is deprecated", DeprecationWarning, stacklevel=2)

        _check_kwargs(kwargs, RTYPE_PROPERTIES + ("description", "__permissions__"))

        self.__dict__.update(kwargs)

    def __str__(self) -> str:
        return f"relation type {self.name!r}"

    @classmethod
    def expand_type_definitions(cls: Type["RelationType"], defined: Defined) -> None:
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name: str = getattr(cls, "name", cls.__name__)

        if cls.__doc__ and not cls.description:
            cls.description = " ".join(cls.__doc__.split())

        if name in defined:
            if defined[name].__class__ is not RelationType:
                raise BadSchemaDefinition(f"duplicated relation type for {name}")

            # relation type created from a relation definition, override it
            all_properties = _RELATION_PROPERTIES() + ("subject", "object")

            _copy_attributes(defined[name], cls, all_properties)

        defined[name] = cls

    @classmethod
    def expand_relation_definitions(
        cls: Type["RelationType"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        name: str = getattr(cls, "name", cls.__name__)

        if getattr(cls, "subject", None) and getattr(cls, "object", None):
            # mypy: "Type[RelationType]" has no attribute "subject"
            # mypy: "Type[RelationType]" has no attribute "object"
            # dynamically set attributes
            relation_definition = RelationDefinition(
                subject=cls.subject,  # type: ignore
                name=name,  # type: ignore
                object=cls.object,  # type: ignore
            )

            relation_definition._add_relations(defined, schema)
            _copy_attributes(cls, relation_definition, _RELATION_DEFINITION_PROPERTIES())


class ComputedRelation(RelationType):
    __permissions__ = MARKER

    def __init__(self, name: Optional[str] = None, rule=None, **kwargs) -> None:
        if rule is not None:
            self.rule = rule

        super().__init__(name, **kwargs)


class RelationDefinition(Definition):
    # FIXME reader magic forbids to define a docstring...
    # """a relation is defined by a name, the entity types that can be
    # subject or object the relation, the cardinality, the constraints
    # and the symmetric property.
    # """

    subject: Union[nullobject, str, List[str], None] = MARKER
    object: Union[nullobject, str, List[str], None] = MARKER
    cardinality: Union[nullobject, str, None] = MARKER
    constraints: List[BaseConstraint] = MARKER  # type: ignore
    symmetric = MARKER
    inlined = MARKER
    formula = MARKER

    def __init__(
        self,
        subject: Optional[Union[str, Tuple[str]]] = None,
        name: Optional[str] = None,
        object: Optional[Union[str, Tuple[str]]] = None,
        package: Optional[str] = None,
        **kwargs,
    ) -> None:
        """kwargs keys must have values in _RELATION_DEFINITION_PROPERTIES()"""
        if subject:
            self.subject = subject
        else:
            self.subject = self.__class__.subject

        if object:
            self.object = object
        else:
            self.object = self.__class__.object

        super().__init__(name)

        global CREATION_RANK
        CREATION_RANK += 1
        self.creation_rank: int = CREATION_RANK

        self.package: str
        if package is not None:
            self.package = package
        elif self.package == "<builtin>":
            self.package = PACKAGE

        if kwargs.pop("meta", None):
            warn("[yams 0.37] meta is deprecated", DeprecationWarning)

        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        _check_kwargs(kwargs, relation_definitions_properties)
        _copy_attributes(attrdict(**kwargs), self, relation_definitions_properties)

        if self.constraints:
            self.constraints = list(self.constraints)

    def __str__(self) -> str:
        return (
            f"relation definition ({self.__dict__['subject']} {self.__dict__['name']} "
            f"{self.__dict__['object']})"
        )

    @classmethod
    def expand_type_definitions(cls: Type["RelationDefinition"], defined: Defined) -> None:
        """schema building step 1:

        register definition objects by adding them to the `defined` dictionnary
        """
        name: str = getattr(cls, "name", cls.__name__)
        relation_type: RelationType = RelationType(name)

        _copy_attributes(cls, relation_type, RTYPE_PROPERTIES)

        if name in defined:
            _copy_attributes(relation_type, defined[name], RTYPE_PROPERTIES)
        else:
            defined[name] = relation_type

        # subject and object in defined's keys are only strings not tuples
        if isinstance(cls.subject, tuple):
            subjects = cls.subject
        else:
            subjects = (cls.subject,)

        if isinstance(cls.object, tuple):
            objects = cls.object
        else:
            objects = (cls.object,)

        for subject in subjects:
            for object in objects:
                key = (subject, name, object)
                if key in defined:
                    raise BadSchemaDefinition(
                        "duplicated relation definition (%s) %s (%s.%s)"
                        % (defined[key], key, cls.__module__, cls.__name__)
                    )
                defined[key] = cls

        # XXX keep this for bw compat
        defined[(cls.subject, name, cls.object)] = cls

    @classmethod
    def expand_relation_definitions(
        cls: Type["RelationDefinition"], defined: Defined, schema: yams_types.Schema
    ) -> None:
        """schema building step 2:

        register all relations definition, expanding wildcard if necessary
        """
        assert cls.subject and cls.object, "%s; check the schema (%s, %s)" % (
            cls,
            cls.subject,
            cls.object,
        )
        cls()._add_relations(defined, schema)

    def _add_relations(self, defined: Defined, schema: yams_types.Schema) -> None:
        name: str = getattr(self, "name", self.__class__.__name__)
        relation_type: Union[autopackage, Definition] = defined[name]
        relation_definitions_properties: Tuple[str, ...] = _RELATION_DEFINITION_PROPERTIES()

        # copy relation definition attributes set on the relation type, beside
        # description
        _copy_attributes(
            relation_type, self, set(relation_definitions_properties) - set(("description",))
        )

        # process default cardinality and constraints if not set yet
        cardinality = self.cardinality
        if cardinality is MARKER:
            if self.object in BASE_TYPES:
                self.cardinality = "?1"
            else:
                self.cardinality = "**"
        else:
            assert isinstance(cardinality, str)
            assert len(cardinality) == 2
            assert cardinality[0] in "1?+*"
            assert cardinality[1] in "1?+*"

        if not self.constraints:
            self.constraints = ()

        relation_schema = schema.relation_schema_for(yams_types.DefinitionName(name))
        if relation_schema.rule:
            raise BadSchemaDefinition(
                'Cannot add relation definition "{}" because an '
                "homonymous computed relation already exists "
                'with rule "{}"'.format(relation_schema.type, relation_schema.rule)
            )

        if self.__permissions__ is MARKER:
            final: bool = next(iter(_actual_types(schema, self.object))) in BASE_TYPES

            if final:
                if self.formula is not MARKER:
                    permissions = DEFAULT_COMPUTED_ATTRPERMS
                else:
                    permissions = DEFAULT_ATTRPERMS
            else:
                permissions = DEFAULT_RELPERMS
        else:
            permissions = self.__permissions__

        for subject in _actual_types(schema, self.subject):
            for object in _actual_types(schema, self.object):
                relation_definition = RelationDefinition(
                    subject, name, object, __permissions__=permissions, package=self.package
                )

                _copy_attributes(self, relation_definition, relation_definitions_properties)
                schema.add_relation_def(relation_definition)


@deprecation.argument_renamed(old_name="etype", new_name="entity_type", version="0.48")
def _actual_types(
    schema: yams_types.Schema, entity_type: Union[str, list, Tuple[Any, Any]]
) -> Union[Generator[Any, Any, None], Tuple[Any], Tuple, List, Any]:
    if entity_type == "*":
        yield from (eschema.type for eschema in schema.entities() if not eschema.final)
    elif isinstance(entity_type, (list, tuple)):
        yield from entity_type
    elif isinstance(entity_type, str):
        yield entity_type
    else:
        raise RuntimeError(
            "Entity types must be strings or list/tuples of strings. "
            'SubjectRelation(Foo) is wrong, SubjectRelation("Foo") is correct. '
            "Hence, %r is not acceptable." % entity_type
        )



# copyright 2019 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with yams. If not, see <https://www.gnu.org/licenses/>.

"""Types declarations for types annotations"""

import decimal
from typing import Union, TYPE_CHECKING, TypeVar, Dict, Tuple, Callable, Any, NewType
import datetime as dt

jsonSerializable = TypeVar("jsonSerializable")
Permission = Tuple[str, ...]
Permissions = Dict[str, Permission]
Checkers = Dict[str, Callable[[Any, Any], bool]]
Converters = Dict[
    str, Callable[[Any], Union[str, bytes, int, float, bool, decimal.Decimal, dt.date, dt.datetime]]
]
DefinitionName = NewType("DefinitionName", str)

# to avoid circular imports
if TYPE_CHECKING:
    from yams import schema, buildobjs

    RdefRdefSchema = Union[schema.RelationDefinitionSchema, buildobjs.RelationDefinition]
    RelationDefinitionSchema = schema.RelationDefinitionSchema
    RelationDefinition = buildobjs.RelationDefinition
    Schema = schema.Schema
    EntitySchema = schema.EntitySchema
    EntityType = buildobjs.EntityType
    ERSchema = schema.ERSchema
    RelationSchema = schema.RelationSchema
    RelationType = buildobjs.RelationType

else:
    RdefRdefSchema = TypeVar("rdef_rdefschema")
    RelationDefinitionSchema = TypeVar("RelationDefinitionSchema")
    RelationDefinition = TypeVar("RelationDefinition")
    Schema = TypeVar("Schema")
    EntitySchema = TypeVar("EntitySchema")
    EntityType = TypeVar("EntityType")
    ERSchema = TypeVar("ERSchema")
    RelationSchema = TypeVar("RelationSchema")
    RelationType = TypeVar("RelationType")
    RelationType = TypeVar("RelationType")



# copyright 2021-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some schema exporters"""

import json
import textwrap
from datetime import datetime, date, time
from io import BytesIO
from abc import ABC, abstractmethod

from yams.types import (
    Schema,
    EntitySchema,
    RelationDefinitionSchema,
)


class SchemaExporter(ABC):
    """Used as base class for classes which exports the schema as a string."""

    @abstractmethod
    def export(self, schema: Schema) -> str:
        raise NotImplementedError()


class JSONSchemaExporter(SchemaExporter):
    """Export schema as a json"""

    def _entity_to_json(self, entity: EntitySchema):
        return {
            "type": entity.type,
            "description": entity.description,
            "final": entity.final,
        }

    def _relation_definition_to_json(
        self, subject: str, object: str, rdef: RelationDefinitionSchema
    ):
        # we don't use the relation definition to find subject and object because
        # in the case of symmetric relation, the definition is the same, since
        # this relation is only defined one time.
        schema_json = {
            "type": rdef.relation_type.type,
            "description": rdef.description or "",
            "final": rdef.final,
            "subject": subject,
            "object": object,
            "cardinality": rdef.cardinality,
            "constraints": self._get_relation_definition_constraints(rdef),
            "options": self._get_relation_definition_options(rdef),
        }
        if hasattr(rdef, "default"):
            default = rdef.default
            if isinstance(default, (datetime, date, time)):
                default = default.isoformat()
            elif isinstance(default, BytesIO):
                default = default.getvalue().decode()
            schema_json["default"] = default
        return schema_json

    def _get_relation_definition_options(self, rdef: RelationDefinitionSchema):
        if rdef.final:
            return dict(
                (option, True)
                for option in ["indexed", "fulltextindexed", "internationalizable"]
                if rdef.get(option, False)
            )
        else:
            options_dict = dict()
            if rdef.relation_type.symmetric:
                options_dict["symmetric"] = True
            if rdef.relation_type.inlined:
                options_dict["inlined"] = True
            if rdef.relation_type.fulltext_container is not None:
                options_dict["fulltext_container"] = rdef.relation_type.fulltext_container
            if rdef.get("composite", False):
                options_dict["composite"] = rdef.get("composite")
            return options_dict

    def _get_relation_definition_constraints(self, rdef: RelationDefinitionSchema) -> list:
        if rdef.constraints is None:
            return []
        return [
            {
                "type": constraint.type(),
                "value": json.loads(constraint.serialize()),
            }
            for constraint in rdef.constraints
        ]

    def export_as_dict(self, schema: Schema) -> dict:
        """
        Export the schema as a Python dict.
        """
        entities_json = [self._entity_to_json(entity) for entity in schema.entities()]
        rdefs_json = sorted(
            (
                self._relation_definition_to_json(subj, obj, rdef)
                for rel in schema.relations()
                # in case of a symmetrical relation between A and B,
                # we'll have in rel.relation_definitions the following keys:
                # {(A, B): rdef, (B, A): rdef}. In this kind of relation,
                # rdef will be the same, with always A as subject and B as
                # object, because it's how the definition is written.
                # That's why we need to keep the keys of the dictionnary.
                for (subj, obj), rdef in rel.relation_definitions.items()
            ),
            key=lambda rdef: rdef["type"],
        )
        return {
            "entities": entities_json,
            "relations_definitions": rdefs_json,
        }

    def export(self, schema: Schema) -> str:
        """
        Export the schema as a JSON dump.
        """
        return json.dumps(
            self.export_as_dict(schema),
            indent=2,
        )


class TypescriptSchemaExporter(SchemaExporter):
    """Export schema as a typescript interface"""

    def __init__(self, interface_name: str):
        super().__init__()
        self._name = interface_name

    def _named_export(self, schema):
        json_exporter = JSONSchemaExporter()
        return f"export interface {self._name} {json_exporter.export(schema)};"

    def _default_export(self, schema):
        json_exporter = JSONSchemaExporter()
        return (
            f"interface InstanceSchema {json_exporter.export(schema)};\n\n"
            "export default InstanceSchema;"
        )

    def _ts_module(self, schema):
        if self._name == "default":
            return self._default_export(schema)
        return self._named_export(schema)

    def export(self, schema: Schema) -> str:
        generation_time = datetime.utcnow().isoformat()
        content = textwrap.wrap(
            "This file is generated. Manual modifications will be overwritten "
            "next time you run `cubicweb-ctl export-schema`.",
            width=80,
            initial_indent="// ",
            subsequent_indent="// ",
        )
        content.append(f"// Generated at: {generation_time}\n")
        content.append(self._ts_module(schema))
        return "\n".join(content)



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <https://www.gnu.org/licenses/>.
"""Write a schema as a dot file."""

import sys
import os.path as osp
from itertools import cycle

from typing import Any, Dict, Generator, Tuple, Set, Sequence, Optional, Union

from logilab.common.graph import DotBackend, GraphGenerator

import yams.types as yams_types


__docformat__: str = "restructuredtext en"

CARD_MAP: Dict[str, str] = {"?": "0..1", "1": "1", "*": "0..n", "+": "1..n"}


class SchemaDotPropsHandler:
    def __init__(self, visitor: "SchemaVisitor") -> None:
        self.visitor = visitor
        # FIXME: colors are arbitrary
        self._colors = cycle(("#aa0000", "#00aa00", "#0000aa", "#000000", "#888888"))
        self.nextcolor = lambda: next(self._colors)

    def node_properties(self, eschema: yams_types.EntitySchema) -> Dict[str, str]:
        """return default DOT drawing options for an entity schema"""
        label = ["{", eschema.type, "|"]

        label.append(
            r"\l".join(
                rel.type
                for rel in eschema.ordered_relations()
                if rel.final and self.visitor.should_display_attr(eschema, rel)
            )
        )

        label.append(r"\l}")  # trailing \l ensure alignement of the last one

        return {
            "label": "".join(label),
            "shape": "record",
            "fontname": "Courier",
            "style": "filled",
        }

    def edge_properties(
        self,
        rschema: Union[yams_types.RelationSchema, None],
        subjnode: yams_types.DefinitionName,
        objnode: yams_types.DefinitionName,
    ) -> Dict[str, str]:
        """return default DOT drawing options for a relation schema"""
        # rschema can be none if the subject is a specialization of the object
        # we get there because we want to draw a specialization arrow anyway
        if rschema is None:
            kwargs = {
                "label": "Parent class",
                "color": "grey",
                "style": "filled",
                "arrowhead": "empty",
            }
        elif rschema.symmetric:
            # symmetric rels are handled differently, let yams decide what's best
            kwargs = {
                "label": rschema.type,
                "color": "#887788",
                "style": "dashed",
                "dir": "both",
                "arrowhead": "normal",
                "arrowtail": "normal",
            }
        else:
            kwargs = {"label": rschema.type, "color": "black", "style": "filled"}
            rdef = rschema.relation_definition(subjnode, objnode)

            # mypy: "RelationDefinitionSchema" has no attribute "composite"
            # this is a dynamically setted attribue using self.__dict__.update(some_dict)
            if rdef.composite == "subject":  # type: ignore
                kwargs["arrowhead"] = "none"
                kwargs["arrowtail"] = "diamond"
            elif rdef.composite == "object":  # type: ignore
                kwargs["arrowhead"] = "diamond"
                kwargs["arrowtail"] = "none"
            else:
                kwargs["arrowhead"] = "normal"
                kwargs["arrowtail"] = "none"

            # UML like cardinalities notation, omitting 1..1

            # mypy: "RelationDefinitionSchema" has no attribute "cardinality"
            # this is a dynamically setted attribue using self.__dict__.update(some_dict)
            if rdef.cardinality[1] != "1":  # type: ignore
                kwargs["taillabel"] = CARD_MAP[rdef.cardinality[1]]  # type: ignore

            if rdef.cardinality[0] != "1":  # type: ignore
                kwargs["headlabel"] = CARD_MAP[rdef.cardinality[0]]  # type: ignore

            kwargs["color"] = self.nextcolor()

        kwargs["fontcolor"] = kwargs["color"]

        # dot label decoration is just awful (1 line underlining the label
        # + 1 line going to the closest edge spline point)
        kwargs["decorate"] = "false"
        # kwargs['labelfloat'] = 'true'

        return kwargs


class SchemaVisitor:
    def __init__(self, skiptypes: Sequence = ()) -> None:
        self._done: Set[Tuple[Any, Any, Any]] = set()
        self.skiptypes: Sequence = skiptypes
        self._nodes: Optional[Set[Tuple[Any, Any]]] = None
        self._edges: Optional[Set[Tuple[Any, Any, Any]]] = None

    def should_display_schema(self, erschema: yams_types.ERSchema) -> bool:
        return not (getattr(erschema, "final", False) or erschema in self.skiptypes)

    def should_display_attr(
        self, eschema: yams_types.EntitySchema, rschema: yams_types.RelationSchema
    ) -> bool:
        return rschema not in self.skiptypes

    def display_rel(
        self,
        rschema: yams_types.RelationSchema,
        setype: yams_types.DefinitionName,
        tetype: yams_types.DefinitionName,
    ) -> bool:
        if (rschema.type, setype, tetype) in self._done:
            return False

        self._done.add((rschema.type, setype, tetype))

        if rschema.symmetric:
            self._done.add((rschema, tetype, setype))

        return True

    def nodes(self) -> Any:
        return self._nodes

    def edges(self) -> Any:
        return self._edges


class FullSchemaVisitor(SchemaVisitor):
    def __init__(self, schema: yams_types.Schema, skiptypes: Sequence = ()) -> None:
        super(FullSchemaVisitor, self).__init__(skiptypes)
        self.schema: yams_types.Schema = schema
        self._eindex: Dict[str, Any] = {
            eschema.type: eschema
            for eschema in schema.entities()
            if self.should_display_schema(eschema)
        }

    def nodes(self) -> Generator[Tuple[str, yams_types.EntitySchema], Any, None]:
        for eschema in sorted(self._eindex.values(), key=lambda x: x.type):
            yield eschema.type, eschema

    def edges(self) -> Generator[Tuple[str, str, Optional[yams_types.RelationSchema]], Any, None]:
        # Entities with inheritance relations.
        for eschema in sorted(self._eindex.values(), key=lambda x: x.type):
            if eschema.specializes():
                yield str(eschema), str(eschema.specializes()), None

        # Subject/object relations.
        for rschema in sorted(self.schema.relations(), key=lambda x: x.type):
            if not self.should_display_schema(rschema):
                continue

            for setype, tetype in sorted(rschema.relation_definitions):
                if not (setype in self._eindex and tetype in self._eindex):
                    continue

                if not self.display_rel(rschema, setype, tetype):
                    continue

                yield setype, tetype, rschema


class OneHopESchemaVisitor(SchemaVisitor):
    def __init__(self, eschema: yams_types.EntitySchema, skiptypes: Sequence = ()) -> None:
        super(OneHopESchemaVisitor, self).__init__(skiptypes)
        nodes = set()
        edges = set()

        nodes.add((eschema.type, eschema))

        for rschema in eschema.subject_relations():
            if not self.should_display_schema(rschema):
                continue

            for teschema in rschema.objects(eschema.type):
                nodes.add((teschema.type, teschema))
                if not self.display_rel(rschema, eschema.type, teschema.type):
                    continue

                edges.add((eschema.type, teschema.type, rschema))

        for rschema in eschema.object_relations():
            if not self.should_display_schema(rschema):
                continue

            for teschema in rschema.subjects(eschema.type):
                nodes.add((teschema.type, teschema))

                if not self.display_rel(rschema, teschema.type, eschema.type):
                    continue

                edges.add((teschema.type, eschema.type, rschema))

        # Inheritance relations.
        if eschema.specializes():
            entity_schema_specialisation = eschema.specializes()
            assert entity_schema_specialisation is not None
            nodes.add((entity_schema_specialisation.type, entity_schema_specialisation))
            # mypy: Argument 1 to "add" of "set" has incompatible type "Tuple[str, Any, None]";
            # mypy: expected "Tuple[str, str, RelationSchema]"
            # situation too complex for mypy to handle
            edges.add((eschema.type, eschema.specializes().type, None))  # type: ignore

        if eschema.specialized_by():
            for pschema in eschema.specialized_by():
                nodes.add((pschema.type, pschema))
                # mypy: Argument 1 to "add" of "set" has incompatible type "Tuple[str, str, None]";
                # mypy: expected "Tuple[str, str, RelationSchema]"
                # situation too complex for mypy to handle
                edges.add((pschema.type, eschema.type, None))  # type: ignore

        self._nodes = nodes
        self._edges = edges


class OneHopRSchemaVisitor(SchemaVisitor):
    def __init__(self, rschema: yams_types.RelationSchema, skiptypes: Sequence = ()) -> None:
        super(OneHopRSchemaVisitor, self).__init__(skiptypes)
        nodes = set()
        edges = set()

        for seschema in rschema.subjects():
            nodes.add((seschema.type, seschema))

            for oeschema in rschema.objects(seschema.type):
                nodes.add((oeschema.type, oeschema))

                if not self.display_rel(rschema, seschema.type, oeschema.type):
                    continue

                edges.add((seschema.type, oeschema.type, rschema))

        self._nodes = nodes
        self._edges = edges


def schema2dot(
    schema: Optional[yams_types.Schema] = None,
    outputfile: Optional[str] = None,
    skiptypes: Sequence = (),
    visitor: Optional[Union[SchemaVisitor, FullSchemaVisitor]] = None,
    prophdlr=None,
    size=None,
) -> Any:
    """write to the output stream a dot graph representing the given schema"""

    # mypy: Incompatible types in assignment (expression has type
    # mypy: "Union[SchemaVisitor, Schema, None]", variable has type "Optional[SchemaVisitor]")
    # mypy bug, Type is Optional[Union[SchemaVisitor, FullSchemaVisitor]] because "schema" will
    # either be None (or False...) or skipped but will never be returned from this expression
    visitor = visitor or (schema and FullSchemaVisitor(schema, skiptypes))  # type: ignore

    assert prophdlr or visitor is not None
    # mypy: Argument 1 to "SchemaDotPropsHandler" has incompatible type "Union[SchemaVisitor,
    # mypy: FullSchemaVisitor, None]"; expected "SchemaVisitor"
    # we check before that either we have a prophdlr or, if not, that visitor is not None
    prophdlr = prophdlr or SchemaDotPropsHandler(visitor)  # type: ignore

    schemaname: str
    if outputfile:
        schemaname = osp.splitext(osp.basename(outputfile))[0]
    else:
        schemaname = "Schema"

    generator = GraphGenerator(
        DotBackend(
            schemaname,
            "BT",
            ratio="compress",
            size=size,
            renderer="dot",
            additionnal_param={
                "overlap": "false",
                "splines": "true",
                # 'polylines':'true',
                "sep": "0.2",
            },
        )
    )

    return generator.generate(visitor, prophdlr, outputfile)


def run() -> None:
    """main routine when schema2dot is used as a script"""
    from yams.reader import SchemaLoader

    loader = SchemaLoader()

    if sys.argv[1:]:
        schema_dir: str = sys.argv[1]
    else:
        print("USAGE: schema2dot SCHEMA_DIR [OUTPUT FILE]")
        sys.exit(1)

    outputfile: Optional[str]

    if len(sys.argv) > 2:
        outputfile = sys.argv[2]
    else:
        outputfile = None

    schema = loader.load([schema_dir], "Test")  # type: ignore # uses old api

    schema2dot(schema, outputfile)


if __name__ == "__main__":
    run()



# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
from logilab.common.testlib import TestCase, unittest_main
from yams.xy import XYRegistry


class XYTC(TestCase):
    def test(self):
        xy = XYRegistry()
        xy.register_prefix("dc", "http://purl.org/dc/elements/1.1/")
        xy.register_prefix("foaf", "http://xmlns.com/foaf/0.1/")
        xy.register_prefix("doap", "http://usefulinc.com/ns/doap#")

        xy.add_equivalence("creation_date", "dc:date")
        xy.add_equivalence("created_by", "dc:creator")
        xy.add_equivalence("description", "dc:description")
        xy.add_equivalence("CWUser", "foaf:Person")
        xy.add_equivalence("CWUser login", "dc:title")
        xy.add_equivalence("CWUser surname", "foaf:Person foaf:name")
        xy.add_equivalence("Project", "doap:Project")
        xy.add_equivalence("Project name", "dc:title")
        xy.add_equivalence("Project name", "doap:Project doap:name")
        xy.add_equivalence("Project creation_date", "doap:Project doap:created")
        xy.add_equivalence("Project created_by CWUser", "doap:Project doap:maintainer foaf:Person")
        xy.add_equivalence("Version", "doap:Version")
        xy.add_equivalence("Version name", "doap:Version doap:name")
        xy.add_equivalence("Version creation_date", "doap:Version doap:created")
        xy.add_equivalence("Version num", "doap:Version doap:revision")

        self.assertEqual(xy.yeq("doap:Project", isentity=True), ["Project"])
        self.assertEqual(xy.yeq("dc:title"), [("CWUser", "login", "*"), ("Project", "name", "*")])
        self.assertEqual(
            xy.yeq("doap:Project doap:maintainer"), [("Project", "created_by", "CWUser")]
        )


if __name__ == "__main__":
    unittest_main()



# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <https://www.gnu.org/licenses/>.
"""Compare two yams schemas

Textual representation of schema are created and standard diff algorithm are
applied.
"""

import subprocess
import tempfile
import os

from typing import (
    Tuple,
    TypeVar,
    Dict,
    Sequence,
    Iterable,
    Any,
    Callable,
    Union,
    Set,
    Type,
    List,
    Optional,
)

import yams.types as yams_types

from yams.interfaces import IConstraint
from yams.constraints import SizeConstraint, UniqueConstraint, StaticVocabularyConstraint
from yams.reader import SchemaLoader
from yams.schema import RelationDefinitionSchema


def quoted(obj: Union[str, Any]) -> str:
    if isinstance(obj, str):
        # NOT repr, because .....
        # strings here turn out as unicode strings (from the repo side)
        # but are (mostly) really str in the schema.py source
        # and we don't want a bogus diff about '...' vs u'...'
        if "'" in obj:
            return f'"{obj}"'
    return f"'{obj}'"


class attr_scope:
    indentation: int = 2
    spacer: str = ""
    nextline: str = ","


class class_scope:
    indentation: int = 1
    spacer: str = " "
    nextline: str = ""


_x_type = TypeVar("_x_type")


def identity(x: _x_type) -> _x_type:
    return x


def format_props(
    props: Dict, scope: Union[Type[attr_scope], Type[class_scope]], ignore: Sequence = ()
) -> str:
    out = []

    for prop, val in sorted(props.items()):
        if prop in ignore:
            continue

        out.append("\t" * scope.indentation + f"{prop}{scope.spacer}={scope.spacer}{val}")

    return (f"{scope.nextline}\n").join(out)


def format_expression(expr) -> str:
    mainvars = expr.mainvars - set("SUXO")

    return "%s(%s%s)" % (
        expr.__class__.__name__,
        quoted(expr.expression),
        ",".join(mainvars) if mainvars else "",
    )


def format_tuple(iterator: Iterable) -> str:
    out = []
    for thing in iterator:
        out.append(thing)

    return f"({','.join(out)})"


def format_perms(
    perms: Dict[str, Iterable], scope: Union[Type[attr_scope], Type[class_scope]], isdefault: bool
) -> str:
    out = []
    for p in ("read", "add", "update", "delete"):
        rule = perms.get(p)
        if not rule:
            continue

        out.append(
            "%s%s: %s"
            % (
                "\t" + "\t" * scope.indentation,
                quoted(p),
                format_tuple(
                    format_expression(r) if not isinstance(r, str) else quoted(r)
                    for r in sorted(rule)
                ),
            )
        )

    return " {%s\n%s\n%s}" % (
        " # default perms" if isdefault else "",
        ",\n".join(out),
        "\t" * scope.indentation,
    )


def format_constraint(cstr: IConstraint) -> str:
    cclass = cstr.__class__.__name__

    if "RQL" in cclass:
        out = [f"{cclass}({quoted(cstr.expression)}"]  # type:ignore[attr-defined]
        mainvars = getattr(cstr, "mainvars", None)

        if mainvars:
            out.append(f", mainvars={quoted(','.join(mainvars))}")

        out.append(")")
    elif "Boundary" in cclass:
        return "%s(%s %s)" % (
            cclass,
            quoted(cstr.operator),  # type:ignore[attr-defined]
            str(cstr.boundary),  # type:ignore[attr-defined]
        )

    else:
        print("formatter: unhandled constraint type", cstr.__class__)

        return str(cstr)

    return "".join(out)


perms_type = TypeVar("perms_type")


def nullhandler(relation, perms: perms_type) -> Tuple[perms_type, bool]:
    return perms, False


_permissionshandlerType = Callable[[Any, perms_type], Tuple[perms_type, bool]]
_ReturnType = Dict[str, Union[str, int, bool, Any, List[Union[str, Any]]]]


def properties_from(
    relation: RelationDefinitionSchema, permissionshandler: _permissionshandlerType = nullhandler
) -> _ReturnType:
    """return a dictionary containing properties of an attribute
    or a relation (if specified with is_relation option)"""
    ret: Dict[str, Any] = {}

    for prop in relation.rproperties():
        if prop == "infered":
            continue

        val = getattr(relation, prop)

        if prop == "permissions":
            val, default = permissionshandler(relation, val)

            if val:
                ret["__permissions__"] = format_perms(val, attr_scope, default)

            continue

        elif prop == "cardinality" and relation.final:
            # attributes: re-sugar cardinality to 'required' prop
            prop = "required"
            val = val[0] == "1"

            if not val:
                continue  # don't emit 'required = False' props

        elif prop in ("indexed", "fulltextindexed", "uid", "internationalizable"):
            if not val:
                continue  # don't emit 'prop = False' props

        elif prop == "constraints":
            other = []
            for constraint in val:
                # first, resugar some classic constraints
                if isinstance(constraint, SizeConstraint):
                    ret["maxsize"] = constraint.max

                elif isinstance(constraint, UniqueConstraint):
                    ret["unique"] = True

                elif isinstance(constraint, StaticVocabularyConstraint):
                    # _x_type is for identity function
                    transform: Callable[[Any], Union[str, _x_type]]

                    if relation.object.type == "String":
                        transform = str
                    else:
                        transform = identity

                    ret["vocabulary"] = [transform(x) for x in constraint.vocabulary()]
                else:
                    other.append(constraint)

            if other:
                ret["constraints"] = f"[{','.join(format_constraint(cstr) for cstr in other)}]"

            continue

        if val is None:
            continue

        if prop == "default":
            if not isinstance(val, bool):
                if not val:
                    continue  # default = ''

        elif prop in ("description", "cardinality", "composite"):
            if val:
                val = quoted(val)
            else:  # let's skip empty strings
                continue

        ret[prop] = val

    return ret


def schema2descr(
    schema: yams_types.Schema, permissionshandler: _permissionshandlerType, ignore: Sequence = ()
) -> str:
    """convert a yams schema into a text description"""
    output = []
    # mypy: Name 'ignore' already defined on line 232
    # but we redefine it here as a set (for optimisation?)
    ignore: Set = set(ignore)  # type: ignore
    multirtypes = []

    for entity in sorted(schema.entities()):
        if entity.final:
            continue

        output.append(f"class {entity.type}(EntityType):\n")
        perms, isdefault = permissionshandler(entity, entity.permissions)
        output.append(f"\t__permissions__ = {format_perms(perms, class_scope, isdefault)}\n\n")

        attributes = [
            (
                attr[0].type,
                attr[1].type,
                properties_from(entity.relation_definition(attr[0].type), permissionshandler),
            )
            for attr in sorted(entity.attribute_definitions())
        ]

        for attr_name, attr_type, attr_props in attributes:
            if attr_name in ignore:
                continue

            output.append(
                "\t%s = %s(\n%s)\n\n"
                % (attr_name, attr_type, format_props(attr_props, attr_scope, ignore))
            )

        subject_relations = [
            (
                rel[0].type,  # rtype
                [r.type for r in rel[1]],  # etype targets
                properties_from(entity.relation_definition(rel[0].type), permissionshandler),
            )  # properties
            for rel in sorted(entity.relation_definitions())
            if rel[2] == "subject"
        ]

        for rtype, targettypes, props in subject_relations:
            targettypes = [etype for etype in targettypes if etype not in ignore]

            if not targettypes:
                continue

            if len(targettypes) == 1:
                ttypes_out = quoted(targettypes[0])
            else:
                multirtypes.append(rtype)
                continue

            output.append(
                "\t%s = SubjectRelation(%s\n%s)\n\n"
                % (rtype, ttypes_out, format_props(props, attr_scope, ignore))
            )

        output.append("\n\n")

    # Handle multi-rdef relation types
    for rtype in sorted(multirtypes):
        items = schema[rtype].rdefs.items()  # type: ignore
        for (subjectschema, objectschema), relation in sorted(items):
            subjetype = subjectschema.type
            objetype = objectschema.type

            output.append(
                "class %s_%s_%s(RelationDefinition):\n"
                % (rtype, subjetype.lower(), objetype.lower())
            )
            output.append(f"\tname = {quoted(rtype)}\n")
            output.append(f"\tsubject = {quoted(subjetype)}\n")
            output.append(f"\tobject = {quoted(objetype)}\n")
            output.append(
                format_props(properties_from(relation, permissionshandler), class_scope, ignore)
            )
            output.append("\n\n")

    return "".join(output)


def schema2file(
    schema: yams_types.Schema,
    output: str,
    permissionshandler: _permissionshandlerType,
    ignore: Sequence = (),
) -> None:
    """Save schema description of schema find
    in directory schema_dir into output file"""
    with open(output, "w") as description_file:
        description_file.write(schema2descr(schema, permissionshandler, ignore))


def schema_diff(
    schema1: yams_types.Schema,
    schema2: yams_types.Schema,
    permissionshandler: _permissionshandlerType = nullhandler,
    diff_tool: Optional[str] = None,
    ignore: Sequence = (),
) -> Tuple[str, str]:
    """schema 1 and 2 are CubicwebSchema
    diff_tool is the name of tool use for file comparison
    """
    tmpdir = tempfile.mkdtemp()
    output1 = os.path.join(tmpdir, "schema1.txt")
    output2 = os.path.join(tmpdir, "schema2.txt")

    schema2file(schema1, output1, permissionshandler, ignore)
    schema2file(schema2, output2, permissionshandler, ignore)

    if diff_tool:
        cmd = f"{diff_tool} {output1} {output2}"
        subprocess.Popen(cmd, shell=True)
    else:
        print(f"description files save in {output1} and {output2}")

    print(output1, output2)

    return output1, output2


if __name__ == "__main__":
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option(
        "-f", "--first-schema", dest="schema1", help="Specify the directory of the first schema"
    )
    parser.add_option(
        "-s", "--second-schema", dest="schema2", help="Specify the directory of the second schema"
    )
    parser.add_option(
        "-d", "--diff-tool", dest="diff_tool", help="Specify the name of the diff tool"
    )
    (options, args) = parser.parse_args()
    if options.schema1 and options.schema2:
        schema1 = SchemaLoader().load([options.schema1])
        schema2 = SchemaLoader().load([options.schema2])
        output1, output2 = schema_diff(schema1, schema2)
        if options.diff_tool:
            cmd = f"{options.diff_tool} {output1} {output2}"
            process = subprocess.Popen(cmd, shell=True)
        else:
            print(f"description files save in {output1} and {output2}")
    else:
        parser.error("An input file name must be specified")



# copyright 2004-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module yams.reader"""

import sys
import os.path as osp
from datetime import datetime, date, time

from logilab.common.testlib import TestCase, unittest_main

from yams import BadSchemaDefinition, DEFAULT_RELPERMS, DEFAULT_ATTRPERMS
from yams.schema import Schema
from yams.reader import SchemaLoader, build_schema_from_namespace
from yams.constraints import StaticVocabularyConstraint, SizeConstraint
from yams.buildobjs import (
    EntityType,
    RelationType,
    RelationDefinition,
    SubjectRelation,
    ComputedRelation,
    Int,
    String,
    Float,
    Datetime,
    Date,
    Boolean,
)

sys.path.insert(0, osp.dirname(__file__))


class SchemaLoaderTC(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.schema = SchemaLoader().load([cls.datadir])

    # test helper functions ###################################################

    def test_get_schema_files(self):
        files = [osp.basename(f) for f in SchemaLoader().get_schema_files(self.datadir)]
        self.assertEqual(files[0], "__init__.py")
        self.assertEqual(
            sorted(files), ["Company.py", "Dates.py", "State.py", "__init__.py", "schema.py"]
        )

    # test load_schema read entity and relation types #######################

    def test_load_schema(self):
        self.assertIsInstance(self.schema, Schema)
        self.assertEqual(self.schema.name, "NoName")
        self.assertListEqual(
            sorted(self.schema.entities()),
            [
                "Affaire",
                "BigInt",
                "Boolean",
                "Bytes",
                "Company",
                "Date",
                "Datetest",
                "Datetime",
                "Decimal",
                "Division",
                "EPermission",
                "Eetype",
                "Employee",
                "Float",
                "Int",
                "Interval",
                "Note",
                "Password",
                "Person",
                "Salaried",
                "Societe",
                "State",
                "String",
                "Subcompany",
                "Subdivision",
                "TZDatetime",
                "TZTime",
                "Time",
                "pkginfo",
            ],
        )
        self.assertListEqual(
            sorted(self.schema.relations()),
            [
                "ad1",
                "ad2",
                "ad3",
                "adel",
                "ass",
                "author",
                "author_email",
                "concerne",
                "copyright",
                "cp",
                "d1",
                "d2",
                "date",
                "datenaiss",
                "debian_handler",
                "description",
                "division_of",
                "dt1",
                "dt2",
                "eid",
                "evaluee",
                "fax",
                "final",
                "initial_state",
                "inline_rel",
                "license",
                "long_desc",
                "mailinglist",
                "meta",
                "modname",
                "name",
                "next_state",
                "nom",
                "obj_wildcard",
                "para",
                "prenom",
                "promo",
                "ref",
                "require_permission",
                "rncs",
                "salary",
                "sexe",
                "short_desc",
                "state_of",
                "subcompany_of",
                "subdivision_of",
                "subj_wildcard",
                "sujet",
                "sym_rel",
                "t1",
                "t2",
                "tel",
                "test",
                "titre",
                "travaille",
                "type",
                "version",
                "ville",
                "web",
                "works_for",
            ],
        )

    def test_entity_schema_for(self):
        entity_schema = self.schema.entity_schema_for("Societe")
        self.assertEqual(entity_schema.description, "")
        self.assertEqual(entity_schema.final, False)
        self.assertListEqual(
            sorted(list(entity_schema.subject_relations.values())),
            [
                "ad1",
                "ad2",
                "ad3",
                "cp",
                "evaluee",
                "fax",
                "nom",
                "rncs",
                "subj_wildcard",
                "tel",
                "ville",
                "web",
            ],
        )
        self.assertListEqual(
            sorted(list(entity_schema.object_relations.values())),
            ["concerne", "obj_wildcard", "travaille"],
        )

        entity_schema = self.schema.entity_schema_for("Eetype")
        self.assertEqual(
            entity_schema.description, "define an entity type, used to build the application schema"
        )
        self.assertEqual(entity_schema.final, False)
        self.assertListEqual(
            sorted(str(r) for r in entity_schema.subject_relations.values()),
            ["description", "final", "initial_state", "meta", "name", "subj_wildcard"],
        )
        self.assertListEqual(
            sorted(str(r) for r in entity_schema.object_relations.values()),
            ["obj_wildcard", "state_of"],
        )

        entity_schema = self.schema.entity_schema_for("Boolean")
        self.assertEqual(entity_schema.description, "")
        self.assertEqual(entity_schema.final, True)
        self.assertListEqual(sorted(list(entity_schema.subject_relations.values())), [])
        self.assertListEqual(
            sorted(list(entity_schema.object_relations.values())), ["final", "meta", "test"]
        )

    # test base entity type's subject relation properties #####################

    def test_indexed(self):
        entity_schema = self.schema.entity_schema_for("Person")
        self.assertFalse(entity_schema.relation_definition("nom").indexed)
        entity_schema = self.schema.entity_schema_for("State")
        self.assertTrue(entity_schema.relation_definition("name").indexed)

    def test_uid(self):
        entity_schema = self.schema.entity_schema_for("State")
        self.assertTrue(entity_schema.relation_definition("eid").uid)
        self.assertFalse(entity_schema.relation_definition("name").uid)

    def test_fulltextindexed(self):
        entity_schema = self.schema.entity_schema_for("Person")
        self.assertRaises(
            AttributeError, getattr, entity_schema.relation_definition("tel"), "fulltextindexed"
        )  # tel is an INT
        self.assertTrue(entity_schema.relation_definition("nom").fulltextindexed)
        self.assertTrue(entity_schema.relation_definition("prenom").fulltextindexed)
        self.assertFalse(entity_schema.relation_definition("sexe").fulltextindexed)
        indexable = sorted(entity_schema.indexable_attributes())
        self.assertEqual(["nom", "prenom", "titre"], indexable)
        self.assertEqual(self.schema.relation_schema_for("works_for").fulltext_container, None)
        self.assertEqual(
            self.schema.relation_schema_for("require_permission").fulltext_container, "subject"
        )
        entity_schema = self.schema.entity_schema_for("Company")
        indexable = sorted(entity_schema.indexable_attributes())
        self.assertEqual([], indexable)
        indexable = sorted(entity_schema.fulltext_relations())
        self.assertEqual([("require_permission", "subject")], indexable)
        containers = sorted(entity_schema.fulltext_containers())
        self.assertEqual([], containers)
        entity_schema = self.schema.entity_schema_for("EPermission")
        indexable = sorted(entity_schema.indexable_attributes())
        self.assertEqual(["name"], indexable)
        indexable = sorted(entity_schema.fulltext_relations())
        self.assertEqual([], indexable)
        containers = sorted(entity_schema.fulltext_containers())
        self.assertEqual([("require_permission", "subject")], containers)

    def test_internationalizable(self):
        entity_schema = self.schema.entity_schema_for("Eetype")
        self.assertTrue(entity_schema.relation_definition("name").internationalizable)
        entity_schema = self.schema.entity_schema_for("State")
        self.assertTrue(entity_schema.relation_definition("name").internationalizable)
        entity_schema = self.schema.entity_schema_for("Societe")
        self.assertFalse(entity_schema.relation_definition("ad1").internationalizable)

    # test advanced entity type's subject relation properties #################

    def test_vocabulary(self):
        entity_schema = self.schema.entity_schema_for("pkginfo")
        self.assertEqual(entity_schema.vocabulary("license"), ("GPL", "ZPL"))
        self.assertEqual(entity_schema.vocabulary("debian_handler"), ("machin", "bidule"))

    def test_default(self):
        entity_schema = self.schema.entity_schema_for("pkginfo")
        self.assertEqual(entity_schema.default("version"), "0.1")
        self.assertEqual(entity_schema.default("license"), None)

    # test relation type properties ###########################################

    def test_relation_schema_for(self):
        relation_schema = self.schema.relation_schema_for("evaluee")
        self.assertEqual(relation_schema.symmetric, False)
        self.assertEqual(relation_schema.description, "")
        self.assertEqual(relation_schema.final, False)
        self.assertListEqual(sorted(relation_schema.subjects()), ["Person", "Salaried", "Societe"])
        self.assertListEqual(sorted(relation_schema.objects()), ["Note"])

        relation_schema = self.schema.relation_schema_for("sym_rel")
        self.assertEqual(relation_schema.symmetric, True)
        self.assertEqual(relation_schema.description, "")
        self.assertEqual(relation_schema.final, False)
        self.assertListEqual(sorted(relation_schema.subjects()), ["Affaire", "Person", "Salaried"])
        self.assertListEqual(sorted(relation_schema.objects()), ["Affaire", "Person", "Salaried"])

        relation_schema = self.schema.relation_schema_for("initial_state")
        self.assertEqual(relation_schema.symmetric, False)
        self.assertEqual(
            relation_schema.description,
            "indicate which state should be used by default when an entity using states is created",
        )
        self.assertEqual(relation_schema.final, False)
        self.assertListEqual(sorted(relation_schema.subjects()), ["Eetype"])
        self.assertListEqual(sorted(relation_schema.objects()), ["State"])

        relation_schema = self.schema.relation_schema_for("name")
        self.assertEqual(relation_schema.symmetric, False)
        self.assertEqual(relation_schema.description, "")
        self.assertEqual(relation_schema.final, True)
        self.assertListEqual(
            sorted(relation_schema.subjects()),
            ["Company", "Division", "EPermission", "Eetype", "State", "Subcompany", "Subdivision"],
        )
        self.assertListEqual(sorted(relation_schema.objects()), ["String"])

    def test_cardinality(self):
        relation_schema = self.schema.relation_schema_for("evaluee")
        self.assertEqual(relation_schema.relation_definition("Person", "Note").cardinality, "**")
        relation_schema = self.schema.relation_schema_for("inline_rel")
        self.assertEqual(relation_schema.relation_definition("Affaire", "Person").cardinality, "?*")
        relation_schema = self.schema.relation_schema_for("initial_state")
        self.assertEqual(relation_schema.relation_definition("Eetype", "State").cardinality, "?*")
        relation_schema = self.schema.relation_schema_for("state_of")
        self.assertEqual(relation_schema.relation_definition("State", "Eetype").cardinality, "+*")
        relation_schema = self.schema.relation_schema_for("name")
        self.assertEqual(relation_schema.relation_definition("State", "String").cardinality, "11")
        relation_schema = self.schema.relation_schema_for("description")
        self.assertEqual(relation_schema.relation_definition("State", "String").cardinality, "?1")

    def test_constraints(self):
        entity_schema = self.schema.entity_schema_for("Person")
        self.assertEqual(len(entity_schema.relation_definition("nom").constraints), 1)
        self.assertEqual(len(entity_schema.relation_definition("promo").constraints), 2)
        self.assertEqual(len(entity_schema.relation_definition("tel").constraints), 0)
        entity_schema = self.schema.entity_schema_for("State")
        self.assertEqual(len(entity_schema.relation_definition("name").constraints), 1)
        self.assertEqual(len(entity_schema.relation_definition("description").constraints), 0)
        entity_schema = self.schema.entity_schema_for("Eetype")
        self.assertEqual(len(entity_schema.relation_definition("name").constraints), 2)

    def test_inlined(self):
        relation_schema = self.schema.relation_schema_for("evaluee")
        self.assertEqual(relation_schema.inlined, False)
        relation_schema = self.schema.relation_schema_for("state_of")
        self.assertEqual(relation_schema.inlined, False)
        relation_schema = self.schema.relation_schema_for("inline_rel")
        self.assertEqual(relation_schema.inlined, True)
        relation_schema = self.schema.relation_schema_for("initial_state")
        self.assertEqual(relation_schema.inlined, True)

    def test_relation_permissions(self):
        relation_schema = self.schema.relation_schema_for("evaluee")
        self.assertEqual(
            relation_schema.relation_definition("Person", "Note").permissions,
            {"read": ("managers",), "delete": ("managers",), "add": ("managers",)},
        )
        self.assertEqual(
            relation_schema.relation_definition("Societe", "Note").permissions,
            {"read": ("managers",), "delete": ("managers",), "add": ("managers",)},
        )
        relation_schema = self.schema.relation_schema_for("concerne")
        self.assertEqual(
            relation_schema.relation_definition("Person", "Affaire").permissions,
            {"read": ("managers",), "delete": ("managers",), "add": ("managers",)},
        )
        self.assertEqual(
            relation_schema.relation_definition("Affaire", "Societe").permissions, DEFAULT_RELPERMS
        )
        relation_schema = self.schema.relation_schema_for("travaille")
        self.assertEqual(
            relation_schema.relation_definition("Person", "Societe").permissions,
            {"read": (), "add": (), "delete": ("managers",)},
        )

    def test_attributes_permissions(self):
        relation_schema = self.schema.relation_schema_for("name")
        self.assertEqual(
            relation_schema.relation_definition("Company", "String").permissions, DEFAULT_ATTRPERMS
        )
        relation_schema = self.schema.relation_schema_for("tel")
        self.assertEqual(
            relation_schema.relation_definition("Person", "Int").permissions,
            {"read": (), "add": ("managers",), "update": ("managers",)},
        )

    def test_entity_permissions(self):
        entity_schema = self.schema.entity_schema_for("State")
        self.assertEqual(
            entity_schema.permissions,
            {
                "read": (
                    "managers",
                    "users",
                    "guests",
                ),
                "add": (
                    "managers",
                    "users",
                ),
                "delete": (
                    "managers",
                    "owners",
                ),
                "update": (
                    "managers",
                    "owners",
                ),
            },
        )

        entity_schema = self.schema.entity_schema_for("Eetype")
        self.assertEqual(
            entity_schema.permissions,
            {
                "read": (
                    "managers",
                    "users",
                    "guests",
                ),
                "add": ("managers",),
                "delete": ("managers",),
                "update": (
                    "managers",
                    "owners",
                ),
            },
        )

        entity_schema = self.schema.entity_schema_for("Person")
        self.assertEqual(
            entity_schema.permissions,
            {
                "read": (
                    "managers",
                    "users",
                    "guests",
                ),
                "add": (
                    "managers",
                    "users",
                ),
                "delete": (
                    "managers",
                    "owners",
                ),
                "update": (
                    "managers",
                    "owners",
                ),
            },
        )


# def test_nonregr_using_tuple_as_relation_target(self):
#          relation_schema = schema.relation_schema_for('see_also')
#          self.assertEqual(relation_schema.symmetric, False)
#          self.assertEqual(relation_schema.description, '')
#          self.assertEqual(relation_schema.final, False)
#          self.assertListEqual(sorted(relation_schema.subjects()), ['Employee'])
#          self.assertListEqual(sorted(relation_schema.objects()), ['Company', 'Division'])
#


class SchemaLoaderModnamesTC(SchemaLoaderTC):
    @classmethod
    def setUpClass(cls):
        def modnames():
            yield ("data", "data.schema")
            for name in ["State", "Dates", "Company", "schema"]:
                yield ("data", ".".join(["data", "schema", name]))

        cls.schema = SchemaLoader().load(modnames())


class BasePerson(EntityType):
    firstname = String(vocabulary=("logilab", "caesium"), maxsize=10)
    lastname = String(constraints=[StaticVocabularyConstraint(["logilab", "caesium"])])


class Person(BasePerson):
    email = String()


class Employee(Person):
    company = String(vocabulary=("logilab", "caesium"))


class Student(Person):
    __specializes_schema__ = True
    college = String()


class X(Student):
    pass


class Foo(EntityType):
    i = Int(required=True, metadata={"name": String()})
    f = Float()
    d = Datetime()


class PySchemaTC(TestCase):
    def test_python_inheritance(self):
        bp = BasePerson()
        p = Person()
        e = Employee()
        self.assertEqual([r.name for r in bp.__relations__], ["firstname", "lastname"])
        self.assertEqual([r.name for r in p.__relations__], ["firstname", "lastname", "email"])
        self.assertEqual(
            [r.name for r in e.__relations__], ["firstname", "lastname", "email", "company"]
        )

    def test_schema_extension(self):
        s = Student()
        self.assertEqual(
            [r.name for r in s.__relations__], ["firstname", "lastname", "email", "college"]
        )
        self.assertEqual(s.specialized_type, "Person")
        x = X()
        self.assertEqual(x.specialized_type, None)

    def test_relationtype(self):
        foo = Foo()
        self.assertEqual(
            ["Int", "String", "Float", "Datetime"], [r.entity_type for r in foo.__relations__]
        )
        self.assertEqual(foo.__relations__[0].cardinality, "11")
        self.assertEqual(foo.__relations__[2].cardinality, "?1")

    def test_maxsize(self):
        bp = BasePerson()

        def maxsize(e):
            for e in e.constraints:
                if isinstance(e, SizeConstraint):
                    return e.max

        self.assertEqual(maxsize(bp.__relations__[0]), 7)
        # self.assertEqual(maxsize(bp.__relations__[1]), 7)
        emp = Employee()
        self.assertEqual(maxsize(emp.__relations__[3]), 7)

    def test_metadata(self):
        foo = Foo()
        self.assertEqual("i_name", foo.__relations__[1].name)

    def test_date_defaults(self):
        _today = date.today()
        _now = datetime.now()
        schema = SchemaLoader().load([self.datadir])
        datetest = schema.entity_schema_for("Datetest")
        dt1 = datetest.default("dt1")
        dt2 = datetest.default("dt2")
        d1 = datetest.default("d1")
        d2 = datetest.default("d2")
        t1 = datetest.default("t1")
        t2 = datetest.default("t2")
        # datetimes
        self.assertIsInstance(dt1, datetime)
        # there's no easy way to test NOW (except monkey patching now() itself)
        delta = dt1 - _now
        self.assertLess(abs(delta.seconds), 5)
        self.assertEqual(date(dt2.year, dt2.month, dt2.day), _today)
        self.assertIsInstance(dt2, datetime)
        # dates
        self.assertEqual(d1, _today)
        self.assertIsInstance(d1, date)
        self.assertEqual(d2, date(2007, 12, 11))
        self.assertIsInstance(d2, date)
        # times
        self.assertEqual(t1, time(8, 40))
        self.assertIsInstance(t1, time)
        self.assertEqual(t2, time(9, 45))
        self.assertIsInstance(t2, time)


class SchemaLoaderTC2(TestCase):
    def tearDown(self):
        SchemaLoader.main_schema_directory = "schema"

    def assertBadInlinedMessage(self, error):
        try:
            self.assertEqual(
                "conflicting values False/True for property inlined of relation 'rel'", str(error)
            )
        except AssertionError:
            self.assertEqual(
                "conflicting values True/False for property inlined of relation 'rel'", str(error)
            )

    def test_bad_relation_type_inlined_conflict1(self):
        class Anentity(EntityType):
            rel = SubjectRelation("Anentity", inlined=True)

        class Anotherentity(EntityType):
            rel = SubjectRelation("Anentity", inlined=False)

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertBadInlinedMessage(cm.exception)

    def test_bad_relation_type_inlined_conflict2(self):
        class Anentity(EntityType):
            rel = SubjectRelation("Anentity", inlined=True)

        class rel(RelationType):
            inlined = False

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertBadInlinedMessage(cm.exception)

    def test_bad_int_size_constraint(self):
        class Entity(EntityType):
            attr = Int(maxsize=40)

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertEqual("size constraint doesn't apply to Int entity type", str(cm.exception))

    def test_bad_vocab_and_size(self):
        class Entity(EntityType):
            attr = String(
                constraints=[StaticVocabularyConstraint(["ab", "abc"]), SizeConstraint(2)]
            )
            # "auto-fixed" when using:
            # vocabulary=['ab', 'abc'], maxsize=1)

        with self.assertRaises(BadSchemaDefinition) as cm:
            schema = build_schema_from_namespace(locals().items())
        self.assertEqual(
            "size constraint set to 2 but vocabulary contains string of greater size",
            str(cm.exception),
        )

    def test_bad_relation_type_relation_definition_conflict(self):
        class foo(RelationType):
            __permissions__ = {"read": ()}

        class Entity(EntityType):
            foo = String(__permissions__={"add": ()})

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(locals().items())
        self.assertEqual(
            (
                "conflicting values {'add': ()}/{'read': ()} "
                "for property __permissions__ of relation 'foo'"
            ),
            str(cm.exception),
        )

    def test_schema(self):
        class Anentity(EntityType):
            rel = SubjectRelation("Anentity", inlined=True)

        class Anotherentity(EntityType):
            rel = SubjectRelation("Anentity")

        class rel(RelationType):
            composite = "subject"
            cardinality = "1*"
            symmetric = True

        schema = build_schema_from_namespace(locals().items())
        self.assertEqual("<builtin>", schema["Anentity"].package)
        relation = schema["rel"]
        self.assertEqual(True, rel.symmetric)
        self.assertEqual(True, rel.inlined)
        self.assertEqual("<builtin>", rel.package)
        relation_definition1 = relation.relation_definition("Anentity", "Anentity")
        self.assertEqual("subject", relation_definition1.composite)
        self.assertEqual("1*", relation_definition1.cardinality)
        self.assertEqual("<builtin>", relation_definition1.package)
        relation_definition2 = relation.relation_definition("Anotherentity", "Anentity")
        self.assertEqual("subject", relation_definition2.composite)
        self.assertEqual("1*", relation_definition2.cardinality)
        self.assertEqual("<builtin>", relation_definition2.package)

    def test_imports(self):
        schema = SchemaLoader().load([self.datadir, self.datadir + "2"], "Test")
        self.assertEqual(
            {"read": (), "add": (), "update": (), "delete": ()}, schema["Affaire"].permissions
        )
        self.assertEqual(
            [str(r) for r, at in schema["MyNote"].attribute_definitions()],
            ["date", "type", "para", "text"],
        )

    def test_duplicated_relation_type(self):
        loader = SchemaLoader()
        loader.defined = {}

        class RT1(RelationType):
            pass

        loader.add_definition(RT1)
        with self.assertRaises(BadSchemaDefinition) as cm:
            loader.add_definition(RT1)
        self.assertEqual(str(cm.exception), "duplicated relation type for RT1")

    def test_relation_type_priority(self):
        loader = SchemaLoader()
        loader.defined = {}

        class RT1Def(RelationDefinition):
            name = "RT1"
            subject = "Whatever"
            object = "Whatever"

        class RT1(RelationType):
            pass

        loader.add_definition(RT1Def)
        loader.add_definition(RT1)
        self.assertEqual(loader.defined["RT1"], RT1)

    def test_unfinalized_manipulation(self):
        class MyEntity(EntityType):
            base_arg_b = String()
            base_arg_a = Boolean()
            base_sub = SubjectRelation("MyOtherEntity")

        class base_obj(RelationDefinition):
            subject = "MyOtherEntity"
            object = "MyEntity"

        class MyOtherEntity(EntityType):
            base_o_obj = SubjectRelation("MyEntity")

        class base_o_sub(RelationDefinition):
            subject = "MyEntity"
            object = "MyOtherEntity"

        MyEntity.add_relation(Date(), name="new_arg_a")
        MyEntity.add_relation(Int(), name="new_arg_b")
        MyEntity.add_relation(SubjectRelation("MyOtherEntity"), name="new_sub")
        MyOtherEntity.add_relation(SubjectRelation("MyEntity"), name="new_o_obj")

        class new_obj(RelationDefinition):
            subject = "MyOtherEntity"
            object = "MyEntity"

        class new_o_sub(RelationDefinition):
            subject = "MyEntity"
            object = "MyOtherEntity"

        schema = build_schema_from_namespace(locals().items())
        self.assertIn("MyEntity", schema.entities())
        my_entity = schema["MyEntity"]
        attributes_def = my_entity.attribute_definitions()
        attributes = sorted(attr[0].type for attr in attributes_def)
        self.assertEqual(["base_arg_a", "base_arg_b", "new_arg_a", "new_arg_b"], attributes)
        relations_def = my_entity.relation_definitions()
        relations = sorted(rel[0].type for rel in relations_def)
        self.assertEqual(
            [
                "base_o_obj",
                "base_o_sub",
                "base_obj",
                "base_sub",
                "new_o_obj",
                "new_o_sub",
                "new_obj",
                "new_sub",
            ],
            relations,
        )

    def test_post_build_callback(self):
        SchemaLoader.main_schema_directory = "schema_post_build_callback"
        schema = SchemaLoader().load([self.datadir], "Test")
        self.assertIn("Toto", schema.entities())


class BuildSchemaTC(TestCase):
    def test_build_schema(self):
        class Question(EntityType):
            number = Int()
            text = String()

        class Form(EntityType):
            title = String()

        class in_form(RelationDefinition):
            subject = "Question"
            object = "Form"
            cardinality = "*1"

        schema = build_schema_from_namespace(vars().items())
        entities = [x for x in schema.entities() if not x.final]
        self.assertCountEqual(["Form", "Question"], entities)
        relations = [x for x in schema.relations() if not x.final]
        self.assertCountEqual(["in_form"], relations)


class ComputedSchemaTC(TestCase):
    def test_computed_schema(self):
        class Societe(EntityType):
            name = String()

        class Employe(EntityType):
            name = String()

        class travaille(RelationDefinition):
            subject = "Employe"
            object = "Societe"

        class est_paye_par(ComputedRelation):
            __permissions__ = {"read": ("managers", "users")}
            rule = "S travaille O"

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual("S travaille O", schema["est_paye_par"].rule)
        self.assertEqual({"read": ("managers", "users")}, schema["est_paye_par"].permissions)

    def test_no_relation_definition_from_computedrelation(self):
        class Personne(EntityType):
            name = String()

        class Mafieu(EntityType):
            nickname = String()

        class est_paye_par(ComputedRelation):
            rule = "S travaille O"

        class est_soudoye_par(RelationDefinition):
            name = "est_paye_par"
            subject = "Personne"
            object = "Mafieu"

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(vars().items())
        self.assertEqual(
            'Cannot add relation definition "est_paye_par" '
            "because an homonymous computed relation already "
            'exists with rule "S travaille O"',
            str(cm.exception),
        )

    def test_invalid_attributes_in_computedrelation(self):
        class Societe(EntityType):
            name = String()

        class Employe(EntityType):
            name = String()

        class travaille(RelationDefinition):
            subject = "Employe"
            object = "Societe"

        class est_paye_par(ComputedRelation):
            rule = "S travaille O"
            inlined = True

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(vars().items())
        self.assertEqual("Computed relation has no inlined attribute", str(cm.exception))

    def test_invalid_permissions_in_computedrelation(self):
        class Societe(EntityType):
            name = String()

        class Employe(EntityType):
            name = String()

        class travaille(RelationDefinition):
            subject = "Employe"
            object = "Societe"

        class est_paye_par(ComputedRelation):
            __permissions__ = {
                "read": ("managers", "users", "guests"),
                "add": ("hacker inside!",),
                "delete": (),
            }
            rule = "S travaille O"

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(vars().items())
        self.assertEqual(
            "Cannot set add/delete permissions on computed relation est_paye_par", str(cm.exception)
        )

    def test_computed_attribute_type(self):
        class Entity(EntityType):
            attr = Int(formula="Any Z WHERE X oattr Z")
            oattr = String()

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual(
            "Any Z WHERE X oattr Z", schema["Entity"].relation_definition("attr").formula
        )
        self.assertIsNone(schema["Entity"].relation_definition("oattr").formula)

    def test_computed_attribute_relation_definition(self):
        class Entity(EntityType):
            oattr = String()

        class attr(RelationDefinition):
            subject = "Entity"
            object = "Int"
            formula = "Any Z WHERE X oattr Z"

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual(
            "Any Z WHERE X oattr Z", schema["Entity"].relation_definition("attr").formula
        )
        self.assertIsNone(schema["Entity"].relation_definition("oattr").formula)

    def test_computed_attribute_perms(self):
        class Entity(EntityType):
            oattr = String()

        class attr(RelationDefinition):
            subject = "Entity"
            object = "Int"
            formula = "Any Z WHERE X oattr Z"

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual(
            {"read": ("managers", "users", "guests"), "update": (), "add": ()},
            schema["attr"].relation_definition("Entity", "Int").permissions,
        )

    def test_cannot_set_addupdate_perms_on_computed_attribute(self):
        class Entity(EntityType):
            oattr = String()

        class attr(RelationDefinition):
            __permissions__ = {
                "read": ("managers", "users", "guests"),
                "add": ("hacker inside!",),
                "update": (),
            }
            subject = "Entity"
            object = "Int"
            formula = "Any Z WHERE X oattr Z"

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(vars().items())
        self.assertEqual(
            "Cannot set add/update permissions on computed attribute Entity.attr[Int]",
            str(cm.exception),
        )

    def test_override_read_perms_on_computed_attribute(self):
        class Entity(EntityType):
            oattr = String()
            subjrel = SubjectRelation("String")

        class attr(RelationDefinition):
            __permissions__ = {"read": ("clows",), "add": (), "update": ()}
            subject = "Entity"
            object = "Int"
            formula = "Any Z WHERE X oattr Z"

        class foo(RelationDefinition):
            subject = "Entity"
            object = "Boolean"

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual(
            {"read": ("clows",), "update": (), "add": ()},
            schema["attr"].relation_definition("Entity", "Int").permissions,
        )
        self.assertEqual(
            DEFAULT_ATTRPERMS, schema["foo"].relation_definition("Entity", "Boolean").permissions
        )

    def test_computed_attribute_subjrel(self):
        class Entity(EntityType):
            oattr = String()
            attr = SubjectRelation("Int", formula="Any Z WHERE X oattr Z")

        schema = build_schema_from_namespace(vars().items())
        self.assertEqual(
            "Any Z WHERE X oattr Z", schema["attr"].relation_definition("Entity", "Int").formula
        )


if __name__ == "__main__":
    unittest_main()



#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,W0152
# copyright 2004-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.
"""Generic Setup script, takes package info from __pkginfo__.py file."""
__docformat__ = "restructuredtext en"

import os.path as osp

from setuptools import setup, find_packages


here = osp.abspath(osp.dirname(__file__))

pkginfo = {}
with open(osp.join(here, "__pkginfo__.py")) as f:
    exec(f.read(), pkginfo)

# Get the long description from the relevant file
with open(osp.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

kwargs = {}
if "subpackage_of" in pkginfo:
    kwargs["namespace_packages"] = [pkginfo["subpackage_of"]]

setup(
    name=pkginfo.get("distname", pkginfo["modname"]),
    version=pkginfo["version"],
    license=pkginfo["license"],
    description=pkginfo["description"],
    long_description=long_description,
    author=pkginfo["author"],
    author_email=pkginfo["author_email"],
    url=pkginfo["web"],
    classifiers=pkginfo.get("classifiers", []),
    packages=find_packages(exclude=["test*"]),
    package_data={"yams": ["py.typed"]},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=pkginfo.get("install_requires"),
    scripts=pkginfo.get("scripts", []),
    ext_modules=pkginfo.get("ext_modules"),
    **kwargs,
)
